import json
import os
from datetime import datetime
from json import JSONDecodeError
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

#from scalene import scalene_profiler

import click

from pyscomotif.constants import (INDEX_MAX_ANGLE_VALUE,
                                    INDEX_MAX_DISTANCE_VALUE)
from pyscomotif.indexing import create_index_folder, update_index_folder
from pyscomotif.motif_search import search_index_for_PDBs_with_similar_motifs
from pyscomotif.RMSD_calculation import calculate_RMSD_between_motif_and_similar_motifs

DISCLAIMER = """
\n
DISCLAIMER: pyScoMotif is free to use for non-commercial purposes. To use pyScoMotif for commercial purposes, please contact us.
pyScoMotif is developed by the 3BioInfo group from the ULB, Belgium.
Reference paper: 'pyScoMotif: Discovery of similar 3D structural motifs across proteins'.
Main developer: Gabriel Cia.
\n
"""

@click.group(help='pyScoMotif: a tool for the discovery of similar 3D structural motifs across proteins.', context_settings={'max_content_width':2000}, epilog=DISCLAIMER)
def command_line_interface() -> None:
    pass


def check_index_path_option(ctx: Any, param: Any, value: Union[None, Path]) -> Path:
    if value:
        return value
    
    default_index_folder_path: Path = ctx.params['database_path'] / 'pyScoMotif_index'
    # In Click, callbacks run AFTER type casting and its associated validation, so we have to re-run the validations.
    default_index_folder_path = click.Path(file_okay=False, dir_okay=True, readable=True, writable=True, path_type=Path).convert(default_index_folder_path, ctx=ctx, param=param)
    
    return default_index_folder_path


@command_line_interface.command(epilog=DISCLAIMER)
@click.argument('database_path', type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True, path_type=Path))
@click.option('--index_path', default=None, show_default=False, type=click.Path(file_okay=False, dir_okay=True, readable=True, writable=True, path_type=Path), 
              callback=check_index_path_option, help="Full path of the directory that will contain the index files. Defaults to <database_path>/pyScoMotif_index.")
@click.option('--pattern', type=str, default='*.pdb', show_default=True,
               help="File extension pattern of the PDB files in the database that the file detection algorithm should match, including compression. Examples: *.pdb, *.pdb.gz, *.ent , etcetc . Note the use of the '*' wildcard. Also note that only simple unix style patterns are accepted, complex regex patterns will fail. To know if your pattern works, test it with pathlib.Path.rglob.")
@click.option('--compression', type=click.Choice(('bz2', 'gz')), default='bz2', show_default=True,
               help="Compression algorithm to use to compress the index files. This has nothing to do with the compression of the PDB files, for that use the '--pattern' option.")
@click.option('--n_cores', default=1, show_default=True,
               help='Number of cores to use.')
def create_index(database_path: Path, index_path: Path, pattern: str, compression: str, n_cores: int) -> None:
    """
    Command to create the index needed for the rapid search of similar 3D structural motifs with pyScoMotif. If the --index_path option is not provided, a directory 
    named 'pyScoMotif_index' will be created in the given database path. Once the indexing is completed, the directory can be renamed or moved 
    somewhere else. To update an existing index with new PDB files, see the 'update-index' command.

    Note that during the indexing, residues without a carbon alpha atom or without any sidechain heavy atoms (except for Glycine) are ignored.

    \b
    Arguments
    ---------
    DATABASE_PATH  Full path of the directory containing the PDB files to index (e.g: /home/user/Downloads/PDB). The file detection algorithm is recursive, 
    so PDB files in subfolders will also be indexed.
    """
    create_index_folder(database_path, pattern, index_path, compression, n_cores)
    return


@command_line_interface.command(epilog=DISCLAIMER)
@click.argument('database_path', type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True, path_type=Path))
@click.argument('index_path', type=click.Path(exists=True, file_okay=False, dir_okay=True, writable=True, readable=True, path_type=Path))
@click.option('--pattern', type=str, default='*.pdb', show_default=True,
               help="File extension pattern of the PDB files in the database that the file detection algorithm should match. Examples: *.pdb, *.pdb.gz, *.ent , etcetc . Note the use of the '*' wildcard. Also note that only simple unix style patterns are accepted, complex regex patterns will fail. To know if your pattern works, test it with pathlib.Path.rglob.")
@click.option('--n_cores', default=1, show_default=True,
               help='Number of cores to use.')
def update_index(database_path: Path, pattern: str, index_path: Path, n_cores: int) -> None:
    """
    Command to update an existing pyScoMotif index with new PDB files. The new PDB files are determined by comparing the name 
    of the PDB files with those that have already been indexed, which are stored in the 'indexed_PDB_files.txt' file.

    \b
    Arguments
    ---------
    DATABASE_PATH  Full path of the directory containing the new PDB files to add to the index (e.g: /home/user/Downloads/PDB). The file detection algorithm is 
    recursive, so PDB files in subfolders will also be indexed. Do not worry if the directory contains PDBs that have already been indexed, they will be ignored.
    
    INDEX_PATH  Full path of the pyScoMotif index directory to update (e.g: /home/user/Downloads/PDB/pyScoMotif_index).
    """
    update_index_folder(database_path, pattern, index_path, n_cores)
    return


def check_motif_argument(ctx: Any, param: Any, value: Tuple[str, ...]) -> Tuple[str, ...]:
    if len(value) < 3:
        raise click.BadParameter(f'The motif of interest must have at least 3 residues, but yours had {len(value)}.')
    return value

def check_results_output_path_option(ctx: Any, param: Any, value: Union[None, Path]) -> Path:
    if value:
        if value.suffix != '.csv':
            raise click.BadParameter("The results_output_path option must be a file path ending in '.csv'")
        return value
        
    else:
        now_timestamp = datetime.now().strftime(f'%H%M%S%d%m%Y') # Format = hours+minutes+seconds+day+month+year
        return Path(os.getcwd()) / f'pyScoMotif_result_{now_timestamp}.csv'

def check_residue_type_policy_option(ctx: Any, param: Any, value: str) -> Union[str, Dict[str,List[str]]]:
    if value in ('strict', 'relaxed', 'fully_relaxed'):
        return value

    # The user gives position specific exchanges in a json formated string, we transform it into a python dictionary 
    try:
        loaded_json: Dict[str,str] = json.loads(value) # loads = load string
        position_specific_exchange: Dict[str, List[str]] = {key:list(value) for key, value in loaded_json.items()} # Ex: {'A43':'KRH'} -> {'A43':['K', 'R', 'H']}
    except JSONDecodeError as exception:
        raise exception

    return position_specific_exchange

@command_line_interface.command(epilog=DISCLAIMER)
@click.argument('index_path', type=click.Path(exists=True, path_type=Path))
@click.argument('PDB_file', type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=Path))
@click.argument('motif', nargs=-1, callback=check_motif_argument) # -1 => Unlimited number of arguments
@click.option('--results_output_path', type=click.Path(path_type=Path), default=None, callback=check_results_output_path_option,
              help='Full path of the csv file where the results will be saved (e.g: /home/user/Downloads/similar_motifs.csv). If not given, the results will be saved in the current working directory in a file named with the current timestamp (e.g: pyScoMotif_result_14230214102022.csv)')
@click.option('--sort_results', is_flag=True, default=True, show_default=True,
               help="Sort the results output file by number of mutations and RMSD. Set to False if results file is too big and doesn't fit into RAM.")
@click.option('--residue_type_policy', default='strict', callback=check_residue_type_policy_option,
               help='''Option to control whether mutated versions of the query motif should also be tested. There are 4 possible values: 1) strict (DEFAULT): only similar motifs with strictly identical residues to the query motif are searched. 2) relaxed: finds PDBs with motifs that are similar but with potential mutated residues by allowing residues to mutate according to the residue group they are part of, such as polar, positively charged, aromatic, etcetc (see our Github for the exact grouping of the residues). 3) fully_relaxed: similar to the relaxed mode, but residues can be mutated to any of the 20 possible residues. 4) custom position specific exchange: specify possible residue mutations for the desired positions in your motif using a JSON formated string (e.g: '{"A1":"KL", "A2":"YW", "A5":"E"}'. Note the use of single and double quotes).\n To control the maximum number of simultaneous mutations, see the max_n_mutated_residues option.''')
@click.option('--max_n_mutated_residues', type=click.IntRange(min=1), default=1, show_default=True,
               help='Maximum number of simultaneous mutations that should be allowed when generating mutated motifs through the residue_type_policy option. For example, when set to 1, versions of the query motif with 1 residue mutated according to the given residue_type_policy will be generated and checked, in addition to the reference query motif itself. Note that very high values for this option will result in longer computation time given all the possible combinations of mutated residues will be generated and searched for in the index. For most use cases, we recommend to use a value between 1 and 3.')
@click.option('--distance_delta_thr', type=click.FloatRange(min=0.0, max=INDEX_MAX_DISTANCE_VALUE, min_open=True, max_open=True), default=2.0, show_default=True,
               help='The distance tolerance (in Å) that should be allowed to consider a pair of residues as similar. A higher value will result in more hits with a higher RMSD, and vice versa. Note that very high values will result in higher RAM usage.')
@click.option('--angle_delta_thr', type=click.FloatRange(min=0.0, max=INDEX_MAX_ANGLE_VALUE, min_open=True, max_open=True), default=30.0, show_default=True,
               help='The angle tolerance (in degrees) that should be allowed to consider a pair of residues as similar. A higher value will result in more hits with a higher RMSD, and vice versa. Note that very high values will result in higher RAM usage.')
@click.option('--RMSD_atoms', type=click.Choice(('CA', 'sidechain', 'CA+sidechain')), default='CA+sidechain', show_default=True,
               help="Determines which coordinates are used to calculate the RMSD. CA = carbon alpha atom coordinate, sidechain = average coordinate of the sidechain's heavy atoms, CA+sidechain = both coordinates.")
@click.option('--RMSD_threshold', type=click.FloatRange(min=0, min_open=True), default=1.5, show_default=True, 
               help='Determines the maximum RMSD of the reported similar motifs in the output csv table. To obtain all hits irrespective of RMSD, set this option to a very high value (e.g 10).')
@click.option('--n_cores', default=1, show_default=True,
               help='Number of cores to use. Note that using high distance_delta_thr and/or angle_delta_thr values in conjuction with many cores could result in very high RAM usage.')
def motif_search(
    index_path: Path, pdb_file: Path, motif: Tuple[str, ...], results_output_path: Path, sort_results: bool,
    residue_type_policy: Union[str, Dict[str,List[str]]], max_n_mutated_residues: int, distance_delta_thr: float, 
    angle_delta_thr: float, rmsd_atoms: str, rmsd_threshold: float, n_cores: int
    ) -> None:
    """
    Command to perform the search of similar 3D structural motifs.
    
    \b
    Arguments
    ---------
    INDEX_PATH  Full path of the pyScoMotif index directory to use for the motif search (e.g: /home/user/Downloads/PDB/pyScoMotif_index).

    PDB_FILE  Full path of the PDB file containing the motif of interest (e.g: /home/user/Downloads/1A2Y.pdb).

    MOTIF  Space separated residue identifiers corresponding to the motif of interest (e.g: A1 A2 A5 A8 A13).
    """
    motif_MST, PDBs_with_similar_motifs = search_index_for_PDBs_with_similar_motifs(
        index_folder_path=index_path, 
        PDB_file=pdb_file, motif=motif, 
        residue_type_policy=residue_type_policy,
        max_n_mutated_residues=max_n_mutated_residues,
        distance_delta_thr=distance_delta_thr, angle_delta_thr=angle_delta_thr,
        n_cores=n_cores
    )

    calculate_RMSD_between_motif_and_similar_motifs(
        motif_MST=motif_MST, 
        PDB_file=pdb_file, 
        PDBs_with_similar_motifs=PDBs_with_similar_motifs, 
        index_folder_path=index_path,
        RMSD_atoms=rmsd_atoms,
        RMSD_threshold=rmsd_threshold,
        n_cores=n_cores,
        results_output_path=results_output_path,
        sort_results=sort_results
    )

    return

if __name__ == '__main__':
    #scalene_profiler.start()
    command_line_interface(max_content_width=2000) # max_content_width=2000 allows help texts to span the entire width of the terminal
    #scalene_profiler.stop()


from collections import defaultdict
import concurrent.futures
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Dict, Iterator, List, Set, Tuple

import pandas as pd
from typing_extensions import TypeAlias

from pyscomotif.constants import (INDEX_ANGLE_BIN_SIZE,
                                  INDEX_DISTANCE_BIN_SIZE,
                                  INDEX_MAX_DISTANCE_VALUE)
from pyscomotif.data_containers import Residue
from pyscomotif.index_folders_and_files import create_index_folder_tree
from pyscomotif.residue_data_dicts import \
    generate_compressed_residue_data_from_PDB_file
from pyscomotif.utils import (
    BoundedProcessPoolExecutor,
    detect_the_compression_algorithm_used_in_the_index,
    generate_all_pairs_of_residues_combinations, get_bin_number,
    get_C_alpha_distance, get_PDB_ID_from_file_path,
    get_sidechain_CMR_distance, get_vector_angle,
    pickle_and_compress_python_object, read_compressed_and_pickled_file)


def get_geometric_attribute(geometric_descriptor: str) -> str:
    return geometric_descriptor.rsplit('_', maxsplit=1)[0] # Ex: C_alpha_distance -> C_alpha

def get_residue_specific_data(residue_data: Dict[str, Residue], residue_name: str) -> Dict[str, Residue]:
    """
    ...
    """
    return {residue_ID:residue for residue_ID, residue in residue_data.items() if residue.resname == residue_name}

index_data_type_alias: TypeAlias = List[Tuple[float, float, float, str, str, str]] # For mypy annotation
def get_index_data(residue_data_of_residue_1: Dict[str, Residue], residue_data_of_residue_2: Dict[str, Residue], PDB_ID: str) -> index_data_type_alias:
    """
    ...
    """
    indexed_residue_pairs: Set[str] = set()
    index_data: index_data_type_alias = []
    for residue_1_ID, residue_1 in residue_data_of_residue_1.items():
        for residue_2_ID, residue_2 in residue_data_of_residue_2.items():
            # In the particular case of pairs of identical residue types (e.g: AA, CC, etc), we have to take care of two edge cases:
            # - 1) Avoid including a residue pair of a residue with itself (e.g (A1C, A1C))
            # - 2) Avoid including twice the same residue pair (e.g (A1C, A5C) and (A5C, A1C)) as they are the same
            if residue_1.resname == residue_2.resname:
                if residue_1_ID == residue_2_ID:
                    continue
                if residue_2_ID+residue_1_ID in indexed_residue_pairs:
                    continue
            
            C_alpha_distance = get_C_alpha_distance(residue_1, residue_2)
            if C_alpha_distance > INDEX_MAX_DISTANCE_VALUE:
                continue

            sidechain_CMR_distance = get_sidechain_CMR_distance(residue_1, residue_2)
            vector_angle = get_vector_angle(residue_1, residue_2)

            index_data.append((round(C_alpha_distance, 2), round(sidechain_CMR_distance, 2), round(vector_angle, 2), residue_1_ID, residue_2_ID, PDB_ID))
            indexed_residue_pairs.add(residue_1_ID+residue_2_ID)

    return index_data

def generate_index_data(residue_data: Dict[str, Residue], PDB_ID:str, residue_pair: str) -> index_data_type_alias:
    """
    ...
    """
    residue_1_name, residue_2_name = residue_pair[0], residue_pair[1]
    
    residue_data_of_residue_1 = get_residue_specific_data(residue_data, residue_1_name)
    residue_data_of_residue_2 = get_residue_specific_data(residue_data, residue_2_name)

    index_data = get_index_data(residue_data_of_residue_1, residue_data_of_residue_2, PDB_ID)

    return index_data

def write_index_data_to_corresponding_bin_files(cumulated_index_data: index_data_type_alias, residue_pair: str, index_folder: Path) -> None:
    """
    ...
    """
    # Residues are written to files corresponding to the combination of the 3 bin values for their geometric descriptors (e.g: AG_4_4_5).
    # The bin is determined by simply calculating the floor division between the metric value and the bin size. Note that the use of floor division 
    # leads to half open intervals.
    # Distance bin example: value = 2.3 -> 2.3 //  1 = 2 = bin number
    # Angle bin example   : value = 160 -> 160 // 20 = 8 = bin number
    index_file_rows_map: Dict[str,List[str]] = defaultdict(list)
    for row in cumulated_index_data:
        C_alpha_distance, sidechain_CMR_distance, vector_angle_bin = row[0], row[1], row[2]

        C_alpha_distance_bin = get_bin_number(C_alpha_distance, INDEX_DISTANCE_BIN_SIZE)
        sidechain_CMR_distance_bin = get_bin_number(sidechain_CMR_distance, INDEX_DISTANCE_BIN_SIZE)
        vector_angle_bin = get_bin_number(vector_angle_bin, INDEX_ANGLE_BIN_SIZE)
        
        index_file = f'{residue_pair}_{C_alpha_distance_bin}_{sidechain_CMR_distance_bin}_{vector_angle_bin}.csv'
        index_file_rows_map[index_file].append(','.join((str(element) for element in row)) + '\n')

    for index_file, rows in index_file_rows_map.items():
        with open(index_folder / index_file, 'a') as file_handle:
            file_handle.write(''.join(rows))

    return

def create_index_files_of_the_residue_pair_combination(residue_pair: str, PDB_files_to_index: List[Path], index_folder_path: Path, compression: str) -> None:
    """
    ...
    """
    compressed_residue_data_folder = index_folder_path / 'residue_data_folder'
    index_folder = index_folder_path / 'index'

    cumulated_index_data: index_data_type_alias = []
    for PDB_file in PDB_files_to_index:
        PDB_ID = get_PDB_ID_from_file_path(PDB_file)
        compressed_residue_data_file_path = compressed_residue_data_folder / f'{PDB_ID}.{compression}' 
        residue_data: Dict[str, Residue] = read_compressed_and_pickled_file(compressed_residue_data_file_path)
        
        index_data = generate_index_data(residue_data, PDB_ID, residue_pair)

        cumulated_index_data.extend(index_data)
        
        # To limit RAM usage even when dealing with hundreds of thousands of PDB files, the index data is periodically written
        # to the corresponding csv files on disk.
        if len(cumulated_index_data) >= 10_000:
            write_index_data_to_corresponding_bin_files(cumulated_index_data, residue_pair, index_folder)
            cumulated_index_data.clear()

    # The above condition never triggers for the last chunk of data, so we need to call the functions one last time.
    write_index_data_to_corresponding_bin_files(cumulated_index_data, residue_pair, index_folder)
    cumulated_index_data.clear()

    # Replace each csv file with a pickled and compressed pandas dataframe. If the pickled and compressed file already
    # exists that means we are in update mode, so we simply read the existing file and concatenate the two dataframes.
    for binned_data_csv_file_path in index_folder.glob(f'{residue_pair}_*.csv'):
        bin_index_data_df = pd.read_csv(
            binned_data_csv_file_path, 
            names=['C_alpha_distance', 'sidechain_CMR_distance', 'vector_angle', 'residue_1','residue_2','PDB_ID'], 
            dtype={'C_alpha_distance':float,'sidechain_CMR_distance':float, 'vector_angle':float, 'residue_1:':str, 'residue_2':str, 'PDB_ID':str},
            engine='c'
        )
        
        output_file_path=binned_data_csv_file_path.with_suffix(f'.{compression}') # Ex: /home/user_name/database_folder/pyScoMotif_index/index/AG_4_4_5.bz2
        if output_file_path.exists(): 
            # Update mode
            bin_index_data_df = pd.concat((bin_index_data_df, read_compressed_and_pickled_file(output_file_path)), axis=0, copy=False) 

        # Replace the csv file with the pickled and compressed pandas dataframe
        binned_data_csv_file_path.unlink()
        pickle_and_compress_python_object(
            python_object=bin_index_data_df,
            output_file_path=output_file_path
        )

    return

def update_file_of_indexed_PDB_files(index_folder: Path, new_PDB_IDs: Iterator[str]) -> None:
    """
    ...
    """
    with open(index_folder / 'indexed_PDB_files.txt', 'a') as file_handle:
        for PDB_ID in new_PDB_IDs:
            file_handle.write(PDB_ID + '\n')
    return

def index_PDB_files(PDB_files_to_index: List[Path], index_folder_path: Path, compression: str, n_cores: int) -> None:
    """
    """
    with BoundedProcessPoolExecutor(max_workers=n_cores, max_submited_tasks=2*n_cores) as executor:
        # We first parse all the PDB files to extract and format the residue data of interest (residue name, C alpha coordinates, etc), 
        # which is then saved as a pickle and compressed file. This is done to avoid parsing anew each PDB file 210 times, instead 
        # we just load these files which contain a python dictionary, which is ~15x faster.
        for PDB_file_path in PDB_files_to_index:
            executor.submit(generate_compressed_residue_data_from_PDB_file, PDB_file_path, index_folder_path, compression)
        
        # Each worker processes 1 of the 210 residue pair combinations. The output is a series of pickled and compressed files for each residue pair that contain 
        # the data of a bin e.g .../pyScoMotif_index/index/AG_4_4_5.bz2 contains the occurences of all AG pairs with a C alpha distance between [4, 5) angstrooms, 
        # a CMR distance between [4, 5) and a vector angle between [100, 120) degrees.
        for combination in generate_all_pairs_of_residues_combinations():
            executor.submit(create_index_files_of_the_residue_pair_combination, combination, PDB_files_to_index, index_folder_path, compression)


    # To allow users to update the index with new structures we need to keep track of the PDB files that have already been indexed.
    update_file_of_indexed_PDB_files(
        index_folder_path, 
        new_PDB_IDs = (get_PDB_ID_from_file_path(PDB_file_path) for PDB_file_path in PDB_files_to_index)
    )
    
    return

def create_index_folder(database_path: Path, pattern: str, index_folder_path: Path, compression: str, n_cores: int) -> None:
    """Indexes all the pairs of residues in the PDB files of the given database."""
    PDB_files_to_index: List[Path] = list(database_path.rglob(pattern)) # rglob = recursively glob through the directory and subdirectory
    if not PDB_files_to_index:
        raise ValueError(f"No files were found that match the pattern {pattern}.")
    
    create_index_folder_tree(index_folder_path)

    index_PDB_files(PDB_files_to_index, index_folder_path, compression, n_cores)
    return

def get_set_of_already_indexed_PDBs(index_folder_path: Path) -> Set[str]:
    """
    Returns the set of PDBs that have already been added to the index.
    """
    with open(index_folder_path / 'indexed_PDB_files.txt', 'rt') as file_handle:
        set_of_already_indexed_PDBs = set(PDB_ID.strip() for PDB_ID in file_handle.readlines())

    return set_of_already_indexed_PDBs

def update_index_folder(database_path: Path, pattern: str, index_folder_path: Path, n_cores: int) -> None:
    """Updates an existing index with new PDB files."""
    set_of_already_indexed_PDBs = get_set_of_already_indexed_PDBs(index_folder_path)
    PDB_files_to_index: List[Path] = []
    for PDB_file in database_path.rglob(pattern):
        PDB_ID = get_PDB_ID_from_file_path(PDB_file)
        if PDB_ID not in set_of_already_indexed_PDBs:
            PDB_files_to_index.append(PDB_file)
    del set_of_already_indexed_PDBs
    
    if not PDB_files_to_index:
        raise ValueError(f"No new files were found that match the pattern {pattern}.")

    compression = detect_the_compression_algorithm_used_in_the_index(index_folder_path)

    index_PDB_files(PDB_files_to_index, index_folder_path, compression, n_cores)

    print(f'{len(PDB_files_to_index)} new structures have been added to the index.')
    return



import shutil
from pathlib import Path
from typing import List, Tuple


def get_index_folders_and_files(index_folder_path: Path) -> Tuple[List[Path], List[Path]]:
    """Paths of the subfolders and files of the index."""
    folders = [ 
        index_folder_path / 'index',
        index_folder_path / 'residue_data_folder' # This is where residue data, extracted from parsing PDB files, is saved as a pickled and compressed file.
    ]
    files = [
        index_folder_path / 'indexed_PDB_files.txt',
    ]
    return (folders, files)

def index_folder_exists(index_folder_path: Path) -> bool:
    """Checks the index folder, its subfolders and files all exist."""
    folders, files = get_index_folders_and_files(index_folder_path)
    return index_folder_path.exists() and all(file.is_file() for file in files) and all(folder.is_dir() for folder in folders)

def create_index_folder_tree(index_folder_path: Path) -> None:
    """Creates the index folder as well as its subfolders and files."""
    if index_folder_path.exists():
        shutil.rmtree(index_folder_path, ignore_errors=True)

    index_folder_path.mkdir()

    folders, files = get_index_folders_and_files(index_folder_path)
    for folder in folders:
        folder.mkdir()
    for file in files:
        file.touch() # Creates an empty file

    return
